// PromptEdit.cpp : implementation file
//

#include "stdafx.h"
#include "PromptEdit.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CPromptEdit

CPromptEdit::CPromptEdit()
{
	menu_entries = 0 , lines_index = 0 , text_length = 0 ;
	pos = 0 ;
}

CPromptEdit::~CPromptEdit()
{
}

BEGIN_MESSAGE_MAP(CPromptEdit, CEdit)
	//{{AFX_MSG_MAP(CPromptEdit)
	ON_WM_KEYDOWN()
	ON_WM_SETFOCUS()
	ON_WM_LBUTTONDOWN()
	ON_WM_RBUTTONDOWN()
	ON_WM_SHOWWINDOW()
	ON_WM_CHAR()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPromptEdit message handlers

void CPromptEdit::SetPrompt( CString pr )
{ 
	prompt_str = pr ;
	limit_text_left = pr.GetLength() ;
	pos = limit_text_left ;
}

void CPromptEdit::Add_Line( CString new_string , BOOL bNew )
{
	CString wnd_text;
	GetWindowText ( wnd_text );
	
	if ( bNew ) wnd_text = wnd_text + "\r\n" + new_string ;
	else wnd_text += new_string ;

	SetWindowText ( wnd_text );
	text_length = wnd_text.GetLength() ;
	
	LineScroll ( GetLineCount(), 0x0 );
}

void CPromptEdit::UpdateCaretPosition()
{
	bKeyFlag = FALSE ;
	///////////////////////////////////////////////////
	// Send Caret position to last character
	for ( int i = 0 ; i < GetLineCount() ; i++ )
		SendMessage( WM_KEYDOWN, VK_DOWN , 0x0 ) ;
	for ( i = 0 ; i < LineLength( GetLineCount() - 1 ) ; i++ )
	SendMessage( WM_KEYDOWN, VK_RIGHT , 0x0 ) ;
	///////////////////////////////////////////////////
	bKeyFlag = TRUE ;
}

void CPromptEdit::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if ( nChar == 0x0D )
	{
		// fixes range limit to enter text
		limit_text_right = pos = limit_text_left ;
		
		// takes last typed command
		cmd_line = ProcessLastCommandLine();

		// update context menu (display it by right clicking the mouse
		// button in the prompt box)
		menu_entries = InsertMenuItem( cmd_line ) ;
		lines_index++ ;
		
		// update prompt box by preserving last k command lines
		// where k is the number of available rows in
		// current edit box
		if ( GetLineCount() >= lines_per_screen )
		{
			CString strBuffer ;
			GetWindowText( strBuffer ) ;
			text_length = strBuffer.GetLength() ;
			strBuffer = strBuffer.Right( text_length - LineLength( 0 ) ) ;
			SetWindowText( strBuffer ) ;
		}
		///////////////////////////////////////////////////
		// display new prompt
		Add_Line( prompt_str , TRUE ) ;
		UpdateCaretPosition() ;
		///////////////////////////////////////////////////
		// Example of Bridge-connection :
		AfxGetMainWnd()->SendMessage( PROMPT_DEFINED_MSG, 0x0 , 0x0 ) ;
		///////////////////////////////////////////////////
	}
	else
	{
		// Ruling key movements
		if ( bKeyFlag )
		{
			if ( nChar == VK_UP ) return ;
			else if ( nChar == VK_DOWN ) return ;
			else if ( nChar == VK_RIGHT )
			{
				if ( pos < limit_text_right ) pos++ ;
				else return ;
			}
			else if ( nChar == VK_BACK ) // it is handled by WM_CHAR
				return ; 
			else if ( nChar == VK_LEFT )
			{
				if ( pos > limit_text_left ) pos-- ;
				else return ;
			}
			else if ( nChar == VK_DELETE )
			{
				if ( pos < limit_text_right ) limit_text_right-- ;
				else return ;
			}
			else
			{
				pos++ ;
				limit_text_right++ ;
			}
		}
	}

	CEdit::OnKeyDown(nChar, nRepCnt, nFlags);
}

CString CPromptEdit::ProcessLastCommandLine()
{
	// Get last command line
	int line_length = LineLength( -1 ) ;
	CString cmd_line ;
	LPTSTR cmd = new char[ line_length ] ;
	
	if ( cmd == NULL ) cmd_line.Empty() ;
	else
	{
		GetLine( GetLineCount() - 1, cmd, line_length ) ;
		cmd_line.FormatMessage( cmd ) ;

		cmd_line = cmd_line.Mid( prompt_str.GetLength(), line_length - prompt_str.GetLength() ) ;
		
		cmd_line.TrimLeft() , cmd_line.TrimRight() ;

		delete cmd ;
	}

	return cmd_line ;
}

int CPromptEdit::InsertMenuItem( CString new_menu_string )
{
	if ( new_menu_string.IsEmpty() ) return menu_entries ;
	//////////////////////////////////////////////////
	// This routine returns how many items are included
	// in the menu, at current time
	//////////////////////////////////////////////////
	// take last initialized string, but if array is
	// still empty the last index is retrieved -1;
	// look at the end of this routine
	int last_indexed_used_string = -1 ;

	for( int i = 0 ; i < lines_per_screen ; i++ )
	{
		if ( p_array_str[ i ] != (CString*)NULL )
			last_indexed_used_string = i ;
		else
		{
			p_array_str[ i ] = new CString ;
			*p_array_str[ i ] = new_menu_string ;
			
			return i + 1 ;
		}
	}
	//////////////////////////////////////////////////
	// All entries are full, so slide them all back for one step 
	// and fill the last entry
	if ( menu_entries == lines_per_screen )
	{
		for( i = 1 ; i < menu_entries ; i++ )
			*p_array_str[ i - 1 ] = *p_array_str[ i ] ;

		*p_array_str[ i - 1 ] = new_menu_string ;
	}
	//////////////////////////////////////////////////
	
	return menu_entries ;
}

void CPromptEdit::OnSetFocus(CWnd* pOldWnd) 
{
	CEdit::OnSetFocus(pOldWnd);
	///////////////////////////////////////////////////
	if ( text_length == 0 )		// full initializations
	{
		// set current prompt : change it as you want
		SetPrompt( "ready >: " ) ;
		// For example here, you may insert a initial text
		// to show as presentation
		Add_Line( "CPromptEdit class @ A.D. 2001", FALSE ) ;
		Add_Line( "coded in MFC/ VC++ 5.0 by Alessandro Rosa", TRUE ) ;
		Add_Line( "e-mail : malilla2000@yahoo.it", TRUE ) ;
		Add_Line( "---------------------------------------------", TRUE ) ;
		Add_Line( "", TRUE ) ;
		///////////////////////////////////////////////
		Add_Line( prompt_str, TRUE ) ;
		///////////////////////////////////////////////
		lines_per_screen = GetMaxLines() ;
		///////////////////////////////////////////////
		// create an array of k CString objects
		// k is the number of available line for
		// the current edit box
		p_array_str = new CString*[ lines_per_screen ] ;
		for( int i = 0 ; i < lines_per_screen ; i++ )
			p_array_str[ i ] = (CString*)NULL ;
		///////////////////////////////////////////////
	}
	
	UpdateCaretPosition();
}

int CPromptEdit::GetMaxLines()
{
	// Get the current metrics to calculate the number
	// of available lines of text, according the
	// current dimensions of the edit box
	CDC* pDC = GetDC() ;
	
	// GET FONT HEIGHT
	TEXTMETRIC tm ;
	pDC->GetOutputTextMetrics( &tm ) ;
	CRect font_rect( 0, 0, tm.tmHeight + tm.tmInternalLeading + tm.tmExternalLeading , tm.tmMaxCharWidth ) ;
	ClientToScreen( font_rect ) ;
	ReleaseDC( pDC ) ;

	// GET EDIT BOX HEIGHT
	CRect client_rect ;
	GetClientRect( client_rect ) ;
	ClientToScreen( client_rect ) ;

	// RETRIEVE HEIGHT PER LINE
	return (int)( (double)client_rect.Height() / (double)font_rect.Height() + 1.0 ) ;
}

void CPromptEdit::OnRButtonDown(UINT nFlags, CPoint point) 
{
	// display context menu
	GetCursorPos( &point ) ;
	ScreenToClient( &point ) ;
	FlyingMenu( point ) ;

	// default handler is no more useful here:
	// I did all the job by myself
//	CEdit::OnRButtonDown(nFlags, point);
}

void CPromptEdit::FlyingMenu( CPoint mouse_point )
{
	// display context menu
	if ( menu_entries == 0 ) return ;
	
	CMenu menu ;
	menu.CreatePopupMenu() ;
	
	CString menu_str ;

	for( int i = 0 ; i < menu_entries ; i++ )
	{
		menu_str = *p_array_str[ i ] ;
		menu_str.TrimLeft() , menu_str.TrimRight() ;
		menu.AppendMenu( MF_STRING, i, menu_str ) ;
	}
	
	ClientToScreen(&mouse_point);
	menu.TrackPopupMenu( TPM_LEFTALIGN | TPM_RIGHTBUTTON,
						 mouse_point.x , mouse_point.y ,
						 this );	
}


void CPromptEdit::OnLButtonDown(UINT nFlags, CPoint point) 
{
	// It sends the caret to last position
	UpdateCaretPosition();
	
	// Intentionally event handler disabled
	//	CEdit::OnLButtonDown(nFlags, point);
}

LRESULT CPromptEdit::DefWindowProc(UINT message, WPARAM wParam, LPARAM lParam) 
{
	// It handles the context menu message sent by each menu item.
	// each menu item is associated with an ID reporting
	// the respective indexed location of the same item.
	// Look at 'FlyingMenu' routine
	if ( message == WM_COMMAND )
	{
		int wID = (int)LOWORD(wParam);
		if ( wID >= 0 && wID < menu_entries ) Menu_Handler( wID ) ;
	}

	return CEdit::DefWindowProc(message, wParam, lParam);
}

void CPromptEdit::Menu_Handler( int menu_item )
{
	///////////////////////////////////////////////////
	// Check if the last row has been inserted
	CString s ;		GetWindowText( s ) ;
	BOOL bNew = ( s.GetLength() > text_length ) ? TRUE : FALSE ;
	///////////////////////////////////////////////////
	// if Yes, a new prompt line is inserted
	if ( bNew ) Add_Line( prompt_str , TRUE ) ;
	// then the command line follows
	Add_Line( *p_array_str[ menu_item ] , FALSE ) ;
	UpdateCaretPosition();
}

void CPromptEdit::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	// detects message of the BACKSPACE key, since
	// not properly handled by WM_KEYDOWN message (!)
	if ( nChar == VK_BACK )
	{
		if ( pos > limit_text_left && pos <= limit_text_right )
		{
			pos-- ;
			limit_text_right-- ;
		}
		else return ;
	}
	
	CEdit::OnChar(nChar, nRepCnt, nFlags);
}
